import React, { ReactNode, forwardRef, useCallback } from 'react';
import {
  Platform,
  StyleProp,
  Text,
  TextStyle,
  requireNativeComponent,
  TextProps,
} from 'react-native';
import { v4 } from 'uuid';
import {
  Highlight,
  HighlightRectsCalculatedNativeEvent,
  RNSelectableTextProps,
  Selection,
  SelectionChangeNativeEvent,
  onWordPressIOSNativeEvent,
} from './types';
import { combineHighlights, mapHighlightsRanges } from './utils';

const RNSelectableText =
  requireNativeComponent<RNSelectableTextProps>('RNSelectableText');

interface HighlightableTextProps extends TextProps {
  value: string;
  style?: StyleProp<TextStyle>;
  highlights: Highlight[];
  highlightStyle?: StyleProp<TextStyle>;
  onLayout?: () => void;
  onDismissSelection?: () => void;
  onSelectionChange?: (selection: Selection) => void;
  onHighlightPress?: (id: string) => void;
  onHighlightRectsCalculated?: (
    event: HighlightRectsCalculatedNativeEvent
  ) => void;
}

export type HighlightableTextRef = {};

export const HighlightableText = forwardRef<
  HighlightableTextRef,
  HighlightableTextProps
>((props, ref) => {
  const {
    value,
    style,
    highlights,
    highlightStyle,
    onLayout,
    onDismissSelection,
    onSelectionChange,
    onHighlightPress,
    onHighlightRectsCalculated,
    ...textProps
  } = props;

  const onTextSelectionChange = useCallback(
    (event: SelectionChangeNativeEvent) => {
      const selection = event.nativeEvent.selection;

      if (selection.start == 0 && selection.end == 0) {
        // ignore
        return;
      }
      
      if (selection.start !== selection.end && selection.rect?.width !== 0) {
        onSelectionChange?.(selection);
        return
      }

      // It seems user has just clicked on a word

      if (highlights.length === 0) {
        onDismissSelection?.();
        return;
      }

      const mergedHighlights = combineHighlights(highlights);

      const highlightInRange = mergedHighlights.find(
        ({ start, end }) => selection.start >= start - 1 && selection.end <= end + 1
      );

      if (highlightInRange) {
        onHighlightPress?.(highlightInRange.id);
      } else {
        onDismissSelection?.();
      }
    },
    [onDismissSelection, onSelectionChange, onHighlightPress, highlights]
  );

  const onWordPressNative = useCallback(
    (event: onWordPressIOSNativeEvent) => {
      if (highlights.length === 0) {
        onDismissSelection?.();
        return;
      }

      const { clickedRangeStart, clickedRangeEnd } = event.nativeEvent;
      const mergedHighlights = combineHighlights(highlights);

      const highlightInRange = mergedHighlights.find(
        ({ start, end }) =>
          clickedRangeStart >= start - 1 && clickedRangeEnd <= end + 1
      );

      if (highlightInRange) {
        onHighlightPress?.(highlightInRange.id);
      } else {
        onDismissSelection?.();
      }
    },
    [highlights, onDismissSelection, onHighlightPress]
  );

  let textValue: ReactNode[] = [value];

  if (highlights.length > 0) {
    textValue = mapHighlightsRanges(value, highlights).map(
      ({ id, isHighlight, text }) => (
        <Text
          key={v4()}
          style={isHighlight ? highlightStyle : {}}
        >
          {text}
        </Text>
      )
    );
  }

  return (
    <RNSelectableText
      {...textProps}
      ref={ref}
      // change key to force re-render whenever highlights change
      key={JSON.stringify(highlights)}
      style={style}
      selectable
      highlights={highlights}
      onWordPress={Platform.OS === 'ios' ? onWordPressNative : () => {}}
      onTextSelectionChange={onTextSelectionChange}
      onHighlightRectsCalculated={onHighlightRectsCalculated}
      onLayout={onLayout}
    >
      <Text>{textValue}</Text>
    </RNSelectableText>
  );
});
